/*
 * Copyright 2008-2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at:
 *
 *     http://aws.amazon.com/apache2.0/
 *
 * or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
 * language governing permissions and limitations under the License.
 */

/**@file */

/**
 * Public interfaces and definitions
 */

#ifndef ION_H_
#define ION_H_

#include "ion_types.h"  /// ion_types.h includes ion_errors.h
#include "ion_string.h"
#include "ion_timestamp.h"
#include "ion_decimal.h"
#include "ion_float.h"
#include "ion_int.h"
#include "ion_collection.h"
#include "ion_symbol_table.h"
#include "ion_stream.h"
#include "ion_reader.h"
#include "ion_writer.h"
#include "ion_catalog.h"
#include "ion_debug.h"
#include "ion_version.h"

#endif
