# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: amazon/ionbenchmark/self_describing_proto.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/amazon/ionbenchmark/self_describing_proto.proto\x1a\x19google/protobuf/any.proto\x1a google/protobuf/descriptor.proto\"z\n\x15SelfDescribingMessage\x12:\n\x0e\x64\x65scriptor_set\x18\x01 \x01(\x0b\x32\".google.protobuf.FileDescriptorSet\x12%\n\x07message\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Anyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'amazon.ionbenchmark.self_describing_proto_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_SELFDESCRIBINGMESSAGE']._serialized_start=112
  _globals['_SELFDESCRIBINGMESSAGE']._serialized_end=234
# @@protoc_insertion_point(module_scope)
