# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
__author__ = 'Amazon.com, Inc.'
__version__ = '0.11.2'

__all__ = [
    'core',
    'exceptions',
    'reader',
    'reader_binary',
    'reader_managed',
    'simple_types',
    'simpleion',
    'symbols',
    'util',
    'writer',
    'writer_binary',
    'writer_binary_raw',
    'writer_binary_raw_fields',
    'writer_buffer',
    'writer_text',
]
